package Align3TP;

import ij.*;
import ij.process.*;

/**
 * ImagePlusPlus holds registration and view affine transformations for
 * an ImagePlus.
 * It also holds a display min and max for the whole stack.
 * And, it holds a weight multiplier but does not use the weight.
 * The display is view[reg[pixels]], i.e. reg is applied first.
 * getSlice() returns an ImageProcessor for x,y,0 of view[reg[pixels]]
 * Instantiating program is responsible for passing a valid ImagePlus.
 * Interpolate selects trilinear interpolation.
 *   The interpolated image goes from (dx, dy, dz) to
 *   (width+dx, height+dy, size+dz) where dx, dy, and dz are < 1.
 *   Hence, last row, column, and slice values are used twice.
 * This class does not currently handle RGB stacks.  It converts them to
 *   byte.  getSlice() would need to be modified to handle 3 values per
 *   pixel.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 9January2002
 *
 * @see Align3TP.Alignment
 * @see	Align3TP.Affine
 * @see	Align3TP.Matrix
 * @see	Align3TP.Vector
 */
public class ImagePlusPlus {
	final static boolean SUBTRACT_MIN_FROM_PIXELS = true;
	final static boolean OPTIMIZE_SPEED = true;
	private ImagePlus imp;
	private Affine view = new Affine(),
					reg = new Affine();
	private double max = java.lang.Double.MIN_VALUE;	// display max
	private double min = java.lang.Double.MAX_VALUE;	// display min
	private double weight = 1.0;

	// pixels[slices+1][height*width]
	//   extra slice helps with interpolation  
	private short[][] pixels;
	private int size, w, h;
	private boolean interpolate = false;

	// common block for getSlice(), findCorners() centeringData(),
	//   getOutD(), getMinD() and getMinDint()
	// minD are the coordinates where the image starts along each axis
	// maxD are the coordinates where the image ends along each axis
	// outD are the output stack dimensions
	private int[] outD = new int[3];
	private double[] minD = new double[3], maxD = new double[3];

	public ImagePlusPlus(ImagePlus imp) {
		this.imp = imp;
		ImageStack stack = imp.getStack();
		size = stack.getSize();
		if(size<1) {	// caller error; should not occur
			IJ.write("Unhandled error: size = "+size+" for stack "+stack);
			return;
		}
		w = imp.getWidth();
		h = imp.getHeight();
		pixels = new short[size+1][];	// extra slice for interpolation
		for(int n=0; n<size; n++) {
			ImageProcessor ip = stack.getProcessor(n+1);
			if(ip==null) {	// caller error; should not occur
				System.err.println("Unhandled error: no slice #"+(n+1)+
									" in "+imp);
				return;
			}
			ip = convertToShortTP(ip);
			max = Math.max(max, ip.getMax());
			min = Math.min(min, ip.getMin());
			pixels[n] = (short[]) ip.getPixels();
		}
		pixels[size] = pixels[size-1];	// extra slice = last slice
		return;
	}

	// converts imageProcessor to short with RGB special case, and
	//   scales values to short
	private ImageProcessor convertToShortTP(ImageProcessor ip) {
		// convertToShort will not resetMinAndMax if already Short
		if(ip instanceof ShortProcessor) {
			ip.resetMinAndMax();
			return ip;
		} if(ip instanceof ColorProcessor)
			ip = ip.convertToByte(false);	// RBG to byte; byte nop
		ip = ip.convertToShort(true);
		return ip;
	}

	// returns current slice
	ImageProcessor getSlice() {
		ImageProcessor ip = imp.getProcessor();
		if(view.isIdentity() && reg.isIdentity()) {
			ImageProcessor ipOut = ip.createProcessor(w, h);
			ipOut.copyBits(ip, 0, 0, Blitter.COPY);
			if(ipOut instanceof ShortProcessor)
				ipOut.resetMinAndMax();
			else
				ipOut = convertToShortTP(ipOut);
			return ipOut;
			}
		findCorners();	// sets outD
		ImageProcessor ipOut = ip.createProcessor(outD[0], outD[1]);
		return getSliceData(ipOut);
	}

	// x[display] = view.transform(reg.transform(x[pixels]))
	// x[pixels] = reg.inverse(view.inverse(x(display)))
	ImageProcessor getSliceData(ImageProcessor ipOut) {
		int outW = ipOut.getWidth(), outH = ipOut.getHeight();
		int[] pixelsOut = new int[outW];
		// find values
		ipOut = convertToShortTP(ipOut);
		int iMin = (int) min;	// make sure this conversion is hoisted
		if(!OPTIMIZE_SPEED) {
			// this loop is logically equivalent to the following (except
			//   for interpolation).  it may be more easily understood,
			//   but is much slower
			for (int j=0; j<outH; j++) {
				for(int i=0; i<outW; i++) {
					double[] xDisp = {(double)i, (double)j, 0.0};
					double[] xPixels = inverse(
										new Vector(xDisp)).value();
					int x = (int) (xPixels[0]+0.5),
						y = (int) (xPixels[1]+0.5),
						n = (int) (xPixels[2]+0.5);
					if(x<0 || y<0 || n<0 || x>=w || y>=h ||n>=size)
						pixelsOut[i] = iMin;
					else
						pixelsOut[i] = pixels[n][y*w+x] & 0xffff;
					if(SUBTRACT_MIN_FROM_PIXELS) {
						pixelsOut[i] -= iMin;
						if(pixelsOut[i] < 0)
							pixelsOut[i] = 0;
					}
				}
				ipOut.putRow(0, j, pixelsOut, outW);
			}
		} else {
			Affine a = reg.inverse().transform(view.inverse());
			Matrix xform = a.getMatrix();
			float[] m = {(float) xform.m[0][0], (float) xform.m[1][0],
							(float) xform.m[2][0]};
			for(int j=0; j<outH; j++) {
				double[] xDisp = {0.0, (double)j, 0.0};
				double[] xPixels = a.transform(
									new Vector(xDisp)).value();
				float[] xPixelsF = {(float) xPixels[0],
									(float) xPixels[1], (float) xPixels[2]};
				// key loop wrt run time
				if(interpolate) {	// trilinear interpolation
					for(int i=0; i<outW; i++) {
						xPixelsF[0] += m[0];
						xPixelsF[1] += m[1];
						xPixelsF[2] += m[2];
						int x = (int) xPixelsF[0],
							y = (int) xPixelsF[1],
							n = (int) xPixelsF[2];
						if(x<0 || y<0 || n<0 || x>=w || y>=h ||n>=size)
							pixelsOut[i] = iMin;
						else {
							int x1, y1;
							x1 = x==w-1 ? x : x+1;	// end of row?
							y1 = y==h-1 ? y : y+1;	// end of column?
							// end of stack dealt with by pixels[size]
							float r0 = Math.max(0F, xPixelsF[0]-(float)x),
									r1 = Math.max(0F, xPixelsF[1]-(float)y),
									r2 = Math.max(0F, xPixelsF[2]-(float)n);
							float r = (
								(1F-r2)*
									((1F-r1)*
										((1F-r0)*pixels[n][y*w+x]
										+ r0*pixels[n][y*w+x1])
									+ r1*
										((1F-r0)*pixels[n][y1*w+x]
										+ r0*pixels[n][y1*w+x1]))
								+ r2*
									((1F-r1)*
										((1F-r0)*pixels[n+1][y*w+x]
										+ r0*pixels[n+1][y*w+x1])
									+ r1*
										((1F-r0)*pixels[n+1][y1*w+x]
										+ r0*pixels[n+1][y1*w+x1]))
							);
							if(r>(float)Short.MAX_VALUE)
								pixelsOut[i] = Short.MAX_VALUE;
							else
								pixelsOut[i] = 0xffff & (int)r;
						}
						if(SUBTRACT_MIN_FROM_PIXELS) {
							pixelsOut[i] -= iMin;
							if(pixelsOut[i] < 0)
								pixelsOut[i] = 0;
						}
					}
				} else {	// nearest neighbor
					for(int i=0; i<outW; i++) {
						xPixelsF[0] += m[0];
						xPixelsF[1] += m[1];
						xPixelsF[2] += m[2];
						int x = (int) (xPixelsF[0]+0.5),
							y = (int) (xPixelsF[1]+0.5),
							n = (int) (xPixelsF[2]+0.5);
						if(x<0 || y<0 || n<0 || x>=w || y>=h ||n>=size)
							pixelsOut[i] = 0;
						else
							pixelsOut[i] = pixels[n][y*w+x] & 0xffff;
						if(SUBTRACT_MIN_FROM_PIXELS) {
							pixelsOut[i] -= iMin;
							if(pixelsOut[i] < 0)
								pixelsOut[i] = 0;
						}
					}
				}
				ipOut.putRow(0, j, pixelsOut, outW);
			}
		}
		return ipOut;
	}

	// maximum values for axes
	int[] getOutD(boolean update) {
		if(update)
			findCorners();
		return outD;
	}

	// minimum values for axes
	double[] getMinD(boolean update) {
		if(update)
			findCorners();
		return minD;
	}

	// minimum values for axes as integer
	int[] getMinDint(boolean update) {
		if(update)
			findCorners();
		int[] minDint = new int[3];
		for(int i=0; i<3; i++)
			minDint[i] = (int) (minD[i]+0.5);
		return minDint;
	}

	// returns offset increments to upper left corner of center image
	double[] centeringData() {
		findCorners();
		double[] addOff = new double[3];
		addOff[0] = -minD[0];		//x
		addOff[1] = -minD[1];		//y
		addOff[2] = -((maxD[2]-minD[2])/2 + minD[2]);	// z
		return addOff;
	}

	// sets up minD, maxD, and outD 
	// minD and maxD are the min and max values for the axes
	//   they are the average of the four smallest or largest corners
	//   e.g. midway on the three lines leading to the extreme corner
	private void findCorners() {
		double[][] corners = {{(double) w-1, (double) h-1, (double) size-1},
						{0.0, (double) h-1, (double) size-1},
						{(double) w-1, 0.0, (double) size-1},
						{0.0, 0.0, (double) size-1},
						{(double) w-1, (double) h-1, 0.0},
						{0.0, (double) h-1, 0.0},
						{(double) w-1, 0.0, 0.0},
						{0.0, 0.0, 0.0} };
		double[][] cXform = new double[3][8];
		double[] c;
		// find the values of the corners in the output space
		for(int i=0; i<8; i++) {
			c = transform(new Vector(corners[i])).value();
			for(int k=0; k<3; k++)
				cXform[k][i] = c[k];
		}
		// the minimum and maximum of the output axes are the average of
		//   the midpoints of the three edges going to the extreme corner
		for(int k=0; k<3; k++) {
			quickSort(cXform[k], 0, 7);
			minD[k] = (3.0*cXform[k][0] + cXform[k][1] +
						cXform[k][2]+ cXform[k][3])/6.0;
			maxD[k] = (3.0*cXform[k][7] + cXform[k][6] +
						cXform[k][5] + cXform[k][4])/6.0;
		}
		for(int k=0; k<2; k++)
			outD[k] = 1 + (int) (maxD[k]+0.5);
		outD[2] = 1 + (int) (maxD[2]-minD[2]+0.5);
		for(int k=0; k<3; k++)
			outD[k] = Math.max(1, outD[k]);
		return;
	}

	private void quickSort(double[] list, int start, int end) {
		if(end<=start) return;
		if(end-start==1 && list[start]<list[end]) return;
		double test = (list[start]+list[end])/2.0;
		int i=start, j=end;
		for(; i<j;) {
			if(list[i]<=test)
				i++;
			else if(list[j]>test)
				j--;
			else {
				double temp=list[i];
				list[i] = list[j];
				list[j] = temp;
				i++;
			}
		}
		quickSort(list, start, i-1);
		quickSort(list, i, end);
		return;
	}

	boolean isAxialView(){
		Matrix xform = view.transform(reg).getMatrix();
		if(xform.m[0][2]==0.0 && xform.m[1][2]==0)
			return true;
		else
			return false;
	}

	Vector transform(Vector in) {
		return view.transform(reg.transform(in));
	}

	Vector inverse(Vector in) {
		return reg.inverse(view.inverse(in));
	}

	// slice is negatvie wrt offset
	double validViewOffset(double testOffset) {
		findCorners();	// sets minD and maxD
		double[] offset = view.getOffset();
		double slice = -(testOffset - offset[2]);
		if(minD[2]<slice)
			if(slice<maxD[2])
				return testOffset;
			else
				return offset[2]-maxD[2];
		else
			return offset[2]-minD[2];
	}

	double getViewOffset() {
		double[] offset = view.getOffset();
		return offset[2];
	}

	void setViewOffset(double slice) {
		double[] offset = view.getOffset();
		offset[2] = slice;
		return;
	}

	void addToView(Affine userInput) {
		view = userInput.transform(view);
		return;
	}

	// userInput is defined in view space.
	// userInput is transformed to reg space only by view.getMatrix(),
	//   (veiw.getOffset() is not used in translation.)
	// transformed input is used to transform reg.
	// userInput.transform(view.transform(reg)
	//   = U*V*R = V*U'*R
	//   => U*V = V*U'
	//   => V[-1]*U*V = V[-1]*V*U' = U'
	void addToReg(Affine userInput) {
		reg = view.inverse().
						transform(userInput).
						transform(view).
						transform(reg);
		return;
	}

	boolean setViewToSlice(int slice) {
		if(slice>=size)
			return false;
		double[] offset = {0.0, 0.0, (double) slice};
		view = new Affine(offset);
		return true;
	}

	boolean setRegToSlice(int slice) {
		if(slice>=size)
			return false;
		double[] offset = {0.0, 0.0, (double) slice};
		reg = new Affine(offset);
		return true;
	}

	void setReg(Affine reg) {
		this.reg = reg;
		return;
	}

	void setView(Affine view) {
		this.view = view;
		return;
	}

	void setWeight(double weight) {
		this.weight = weight;
		return;
	}

	Affine getReg() {
		return reg;
	}

	Affine getView() {
		return view;
	}

	double getWeight() {
		return weight;
	}

	double getMax() {
		return max;
	}

	double getMin() {
		return min;
	}

	boolean getInterpolate() {
		return interpolate;
	}

	void setInterpolate(boolean interpolate) {
		this.interpolate = interpolate;
		return;
	}
}
