import java.awt.*;
import java.awt.event.*;
import java.io.*;
import ij.plugin.frame.*;
import ij.*;
import ij.gui.*;

import Align3TP.Alignment;

/**
 * This ImageJ plugin allows the user to align two source stacks.
 * The source stacks are the first and second open stacks in the "Window"
 *   menu.
 * The alignment may optionally be applied to a third stack, the current
 *   stack.  For example, a PET transmission scan could be aligned with a
 *   CT, and the alignment applied to a PET emission scan.
 *
 * An object of Align3_TP class is a frame with several buttons.
 *   Pushing a button starts a new thread to perform the task.
 *
 * This plug-in does not handle color information.  RGB data is converted
 *   to byte.  Note however that output stacks can be colored and then
 *   combined using ImageJ options.
 *
 * This plugin should work with Java 1.1.  There are many attractive
 *   functions in Platform 2, v1.3 for image processing; however, Wayne
 *   Rasband, has provided a great deal of functionality under 1.1
 *   including a beautifully structured API.  Using only 1.1 features
 *   allows more compatibility with legacy systems.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 7January2002
 *
 * @see align3TP.Alignment
 */
public class Align3_TP extends PlugInFrame implements ActionListener {
	final static String[] option = {
		"Start over",			// 0
		"Layout images",		// 1
		"Undo",					// 2
		"Hello",				// 3
		"Align slices",			// 4
		"Match ROIs",			// 5
		"Shear horizontal",		// 6
		"Shear vertical",		// 7
		"Mouse translate",		// 8
		"Mouse rotate",			// 9
		"Mouse resize",			// 10
		"Mouse cancel",			// 11
		"Resample along line",	// 12
		"Axial (x,y)",			// 13
		"Coronal (x,z)",		// 14
		"Sagital (y,z)",		// 15
		"Scale stacks (z)",		// 16
		"Rotate",				// 17
		"Mirror horizontal",	// 18
		"Mirror vertical",		// 19
		"Increment slice",		// 20
		"Decrement slice",		// 21
		"Goto slice",			// 22
		"Goto again",			// 23
		"About Align3_TP",		// 24
		"Output help",			// 25
		"Output",				// 26
		"Show affine"			// 27
	};

	private static Frame instance;
	private Alignment a;		// The Alignment class is the work horse
	private Panel panel;
	private Button[] button = new Button[option.length];

	public Align3_TP() {
		super("Align Stacks");

		if (instance!=null) {	// another instance running
			instance.toFront();
			return;
		}
		instance = this;
		IJ.register(Align3_TP.class);	// remember instance

		// setup Alignment class and images and windows
		a = new Alignment();
		if(a.setup()==false) {
			instance = null;
			return;
		}

		// set up Panel with buttons
		setLayout(new FlowLayout());
		panel = new Panel();
		panel.setLayout(new GridLayout(7, 4, 5, 5));
		for(int i=0; i<option.length; i++){
			button[i] = addButton(option[i]);
		}
		add(panel);
		pack();
		GUI.center(this);
		show();
		return;
	}

	// make button with ActionListener
	private Button addButton(String label) {
		Button b = new Button(label);
		b.addActionListener(this);
		panel.add(b);
		return b;
	}

	public void showAbout() {
		a.showAbout();
		return;
	}

	// button press dispatching point
	public void actionPerformed(ActionEvent e) {
		String label = e.getActionCommand();
		if (label==null) return;
		for(int index=0; index<option.length; index++) {
			if (label.equals(option[index])) {
				new RunOption(index, a, this);	// start thread
				break;
			}
		}
		return;
	}

	public void processWindowEvent(WindowEvent e) {
		super.processWindowEvent(e);
		if (e.getID()==WindowEvent.WINDOW_CLOSING) {
			instance = null;
			a.unlock();
		}
		return;
	}

	public void dispose() {
		a.closeWindows();
		super.dispose();
		return;
	}

// end of Align3_TP main frame.  The following class has been made
//   inner classes of Align3_TP so that its name will be unique wrt
//   other plugins.


/**
 * An object of this class is instantiated in Align3_TP.actionPerformed
 * After a button is pushed this class provides a new service thread
 * Between threads, user may do other ImageJ commands
 */
public class RunOption extends Thread {
	private Alignment a;
	private int index;
	private Align3_TP align3TP;

	public RunOption(int index,  Alignment a, Align3_TP align3TP) {
		super("Align3TP Thread");
		this.index = index;
		this.a = a;
		this.align3TP = align3TP;
		setPriority(Math.max(getPriority()-2, MIN_PRIORITY));
		start();
		return;
	}

	public void run() {
		if (!a.setup()) return;		// verify image window display
		if(index!=11)
			if (!a.lock()) return;	// lock images
		try {
			if(a.doCommand(index, align3TP))
				endCommand();
			}
		catch(OutOfMemoryError e) {
			IJ.outOfMemory("Align3_TP");
		} catch(Exception e) {
			CharArrayWriter caw = new CharArrayWriter();
			PrintWriter pw = new PrintWriter(caw);
			e.printStackTrace(pw);
			IJ.write(caw.toString());
			IJ.showStatus("");
		}
		return;
	}

	void endCommand() {
		// bring main frame to front with focus at end of thread
		a.unlock();
		align3TP.toFront();
		align3TP.requestFocus();
		return;
	}

}	// end of RunOption class


}	// end of the Align3_TP class

